/*Screen {*/
/*    align: center middle;*/
/*}*/

/*DataPreview {*/
/*    height: 100%;*/
/*    !*height: 100%;*!*/
/*    !*padding: 1 2;*!*/
/*    background: $panel;*/
/*    color: $text;*/
/*    !*overflow: scroll;*!*/
/*    !*border: $secondary tall;*!*/
/*    !*content-align: center middle;*!*/
/*}*/

DataViewPane {
    /*overflow: scroll;*/
    /*background: blue;*/
    /*border: $secondary tall;*/
}
DataViewControl {
    dock: bottom;
    /*border: $secondary tall;*/
    padding: 1 0;
    /*background: red;*/
    /*height: 5;*/
}
